#include "bmp.h"
#include "gbio.h"
#include "translate.h"

//*******************************************

#define	low_byte(w)	((byte)  ((w)&0x00ff)    )
#define	high_byte(w)	((byte) (((w)&0xff00)>>8))
#define	make_word(a,b)	(((word)a) + (((word)b) << 8))

#define	BFT_BMAP	0x4d42
#define	BFT_BITMAPARRAY	0x4142
#define	BCA_UNCOMP	0x00000000L
#define	BCA_RLE8	0x00000001L
#define	BCA_RLE4	0x00000002L
#define	BCA_HUFFFMAN1D	0x00000003L
#define	BCA_RLE24	0x00000004L
#define	MSWCC_EOL	0
#define	MSWCC_EOB	1
#define	MSWCC_DELTA	2

struct BMP_PRIV
{
	BOOL windows;
	dword cbFix;
	dword ulCompression;
	dword cclrUsed;
	dword offBits;
	dword base;
};

static BOOLEAN read_word(FILE_HANDLE fd, word *w)
{
	byte low = 0, high = 0;
	
	if ( file_read(fd, &low, 1) != 1 )
		return FALSE;
	if ( file_read(fd, &high, 1) != 1 )
		return FALSE;
	*w = (word) (low + ((word) high << 8));
	return TRUE;
}

static BOOLEAN read_dword(FILE_HANDLE fd, dword *d)
{
	word low, high;
	if ( !read_word(fd, &low) )
		return FALSE;
	if ( !read_word(fd, &high) )
		return FALSE;
	*d = low + ((dword) high << 16);
	return TRUE;
}

static BOOLEAN write_word(FILE_HANDLE fd, word w)
{
	byte	low  = (byte) w;
	byte	high = (byte) (w >> 8);
	
	file_write(fd, &low, 1);
	file_write(fd, &high, 1);
	return TRUE;
}

static BOOLEAN write_dword(FILE_HANDLE fd, dword d)
{
	write_word(fd, (word) d);
	write_word(fd, (word) (d >> 16));
	return TRUE;
}

//*******************************************************************

HBITMAP _LoadBMPFile(FILE_HANDLE file, ImageStruct* pImage)
{
	void* HeaderInfo;
	HBITMAP Bitmap;

	if (ReadBMPHeader(file,pImage,&HeaderInfo) != GBM_ERR_OK)
	{
		file_close(file);
		return NULL;
	}

	if (pImage->iBitPerPixel > 8)
	{
		BITMAPINFO bmInfo;
		unsigned char* pBits;
		BITMAPINFOHEADER* pBitmapHeader = &bmInfo.bmiHeader;
		pBitmapHeader->biBitCount = pImage->iBitPerPixel;
		pBitmapHeader->biClrImportant = 0;
		pBitmapHeader->biClrUsed = 0;
		pBitmapHeader->biCompression = BI_RGB;
		pBitmapHeader->biHeight = (int)pImage->iHeight;
		pBitmapHeader->biPlanes = 1;
		pBitmapHeader->biSize = sizeof(BITMAPINFOHEADER);
		pBitmapHeader->biSizeImage = 0;
		pBitmapHeader->biWidth = pImage->iWidth;
		pBitmapHeader->biXPelsPerMeter = 0;
		pBitmapHeader->biYPelsPerMeter = 0;
		HDC dc = GetDC(0);
		Bitmap=CreateDIBSection(dc,&bmInfo,DIB_RGB_COLORS,(void**)&pBits,NULL,0);
		if (!Bitmap)
		{
			file_close(file);
			return NULL;
		}
		if (ReadBMPData(file, pImage, HeaderInfo,pBits) != GBM_ERR_OK)
		{
			file_close(file);
			DeleteObject(Bitmap);
			return NULL;
		}
	}
	else
	{
		PALETTE pal;
		if (ReadBMPPalette(file,pImage,HeaderInfo,&pal) != GBM_ERR_OK)
		{
			file_close(file);
			return NULL;
		}

		int ColorCount = 1 << pImage->iBitPerPixel;
		BITMAPINFO* pbmInfo = (BITMAPINFO*)malloc(sizeof(BITMAPINFO) + (ColorCount - 1) * sizeof(RGBQUAD));
		if (!pbmInfo)
		{
			file_close(file);
			return NULL;
		}
		unsigned char* pBits;
		BITMAPINFOHEADER* pBitmapHeader = &pbmInfo->bmiHeader;
		pBitmapHeader->biBitCount = pImage->iBitPerPixel;
		pBitmapHeader->biClrImportant = 0;
		pBitmapHeader->biClrUsed = 0;
		pBitmapHeader->biCompression = BI_RGB;
		pBitmapHeader->biHeight = (int)pImage->iHeight;
		pBitmapHeader->biPlanes = 1;
		pBitmapHeader->biSize = sizeof(BITMAPINFOHEADER);
		pBitmapHeader->biSizeImage = 0;
		pBitmapHeader->biWidth = pImage->iWidth;
		pBitmapHeader->biXPelsPerMeter = 0;
		pBitmapHeader->biYPelsPerMeter = 0;
		for (int i = 0; i < ColorCount; i++)
		{
			pbmInfo->bmiColors[i].rgbBlue = pal.colors[i].r;
			pbmInfo->bmiColors[i].rgbGreen = pal.colors[i].g;
			pbmInfo->bmiColors[i].rgbRed = pal.colors[i].b;
			pbmInfo->bmiColors[i].rgbReserved = 0;
		}
		HDC dc = GetDC(0);
		Bitmap=CreateDIBSection(dc, pbmInfo, DIB_RGB_COLORS, (void**)&pBits, NULL, 0);
		if (!Bitmap)
		{
			free(pbmInfo);
			file_close(file);
			return NULL;
		}
		if (ReadBMPData(file, pImage, HeaderInfo,pBits) != GBM_ERR_OK)
		{
			file_close(file);
			DeleteObject(Bitmap);
			return NULL;
		}
		free(pbmInfo);
	}

	file_close(file);
	return Bitmap;
}

DLL_FUNC HBITMAP LoadBMPFileA(char* filename, ImageStruct* pImage)
{
	FILE_HANDLE file = file_openA(filename);
	if (file == (FILE_HANDLE)-1)
		return NULL;
	
	return _LoadBMPFile(file, pImage);
}

DLL_FUNC HBITMAP LoadBMPFileA(wchar_t* filename, ImageStruct* pImage)
{
	FILE_HANDLE file = file_openW(filename);
	if (file == (FILE_HANDLE)-1)
		return NULL;

	return _LoadBMPFile(file, pImage);
}

DLL_FUNC int ReadBMPHeader(FILE_HANDLE file, ImageStruct* pImage, void** HeaderInfo)
{
	word usType, xHotspot, yHotspot;
	dword cbSize, offBits, cbFix;

	if (!file || !pImage || ! HeaderInfo)
	{
		return GBM_ERR_BAD_ARG;
	}
	
	BMP_PRIV* bmp_priv = new BMP_PRIV;
	if (!bmp_priv)
		return GBM_ERR_MEM;

	*HeaderInfo = NULL;

	if ( !read_word(file, &usType) )
	{
		delete bmp_priv;
		return GBM_ERR_READ;
	}
	
	if ( usType != BFT_BMAP )
	{
		delete bmp_priv;
		return GBM_ERR_BAD_MAGIC;
	}
	
	bmp_priv->base = (dword) ( file_seek(file, 0L, SEEK_CUR) - 2L );

	if ( !read_dword(file, &cbSize) )
	{
		delete bmp_priv;
		return GBM_ERR_READ;
	}
	if ( !read_word(file, &xHotspot) )
	{
		delete bmp_priv;
		return GBM_ERR_READ;
	}
	if ( !read_word(file, &yHotspot) )
	{
		delete bmp_priv;
		return GBM_ERR_READ;
	}
	if ( !read_dword(file, &offBits) )
	{
		delete bmp_priv;
		return GBM_ERR_READ;
	}
	if ( !read_dword(file, &cbFix) )
	{
		delete bmp_priv;
		return GBM_ERR_READ;
	}
	
	bmp_priv->offBits = offBits;
	
	if ( cbFix == 12 )
	{
		word cx, cy, cPlanes, cBitCount;
		
		if ( !read_word(file, &cx) )
		{
			delete bmp_priv;
			return GBM_ERR_READ;
		}
		if ( !read_word(file, &cy) )
		{
			delete bmp_priv;
			return GBM_ERR_READ;
		}
		if ( !read_word(file, &cPlanes) )
		{
			delete bmp_priv;
			return GBM_ERR_READ;
		}
		if ( !read_word(file, &cBitCount) )
		{
			delete bmp_priv;
			return GBM_ERR_READ;
		}
		
		if ( cx == 0 || cy == 0 )
		{
			delete bmp_priv;
			return GBM_ERR_BAD_SIZE;
		}
		if ( cPlanes != 1 )
		{
			delete bmp_priv;
			return GBM_ERR_BMP_PLANES;
		}
		if ( cBitCount != 1 && cBitCount != 4 && cBitCount != 8 && cBitCount != 24 )
		{
			delete bmp_priv;
			return GBM_ERR_BMP_BITCOUNT;
		}
		
		pImage->iWidth  = (int) cx;
		pImage->iHeight = (int) cy;
		pImage->iBitPerPixel = (int) cBitCount;
		
		bmp_priv->windows = FALSE;
	}
	else if ( cbFix >= 16 && cbFix <= 64 &&
		((cbFix & 3) == 0 || cbFix == 42 || cbFix == 46) )
	{
		word cPlanes, cBitCount, usUnits, usReserved, usRecording, usRendering;
		dword ulWidth, ulHeight, ulCompression;
		dword ulSizeImage, ulXPelsPerMeter, ulYPelsPerMeter;
		dword cclrUsed, cclrImportant, cSize1, cSize2, ulColorEncoding, ulIdentifier;
		BOOLEAN ok;
		
		ok  = read_dword(file, &ulWidth);
		ok &= read_dword(file, &ulHeight);
		ok &= read_word(file, &cPlanes);
		ok &= read_word(file, &cBitCount);
		if ( cbFix > 16 )
			ok &= read_dword(file, &ulCompression);
		else
			ulCompression = BCA_UNCOMP;
		if ( cbFix > 20 )
			ok &= read_dword(file, &ulSizeImage);
		if ( cbFix > 24 )
			ok &= read_dword(file, &ulXPelsPerMeter);
		if ( cbFix > 28 )
			ok &= read_dword(file, &ulYPelsPerMeter);
		if ( cbFix > 32 )
			ok &= read_dword(file, &cclrUsed);
		else
			cclrUsed = ( (dword)1 << cBitCount );
		if ( cBitCount != 24 && cclrUsed == 0 )
			cclrUsed = ( (dword)1 << cBitCount );
		
		/* Protect against badly written bitmaps! */
		if ( cclrUsed > ( (dword)1 << cBitCount ) )
			cclrUsed = ( (dword)1 << cBitCount );
		
		if ( cbFix > 36 )
			ok &= read_dword(file, &cclrImportant);
		if ( cbFix > 40 )
			ok &= read_word(file, &usUnits);
		if ( cbFix > 42 )
			ok &= read_word(file, &usReserved);
		if ( cbFix > 44 )
			ok &= read_word(file, &usRecording);
		if ( cbFix > 46 )
			ok &= read_word(file, &usRendering);
		if ( cbFix > 48 )
			ok &= read_dword(file, &cSize1);
		if ( cbFix > 52 )
			ok &= read_dword(file, &cSize2);
		if ( cbFix > 56 )
			ok &= read_dword(file, &ulColorEncoding);
		if ( cbFix > 60 )
			ok &= read_dword(file, &ulIdentifier);
		
		if ( !ok )
		{
			delete bmp_priv;
			return GBM_ERR_READ;
		}
		
		if ( ulWidth == 0L || ulHeight == 0L )
		{
			delete bmp_priv;
			return GBM_ERR_BAD_SIZE;
		}
		if ( cPlanes != 1 )
		{
			delete bmp_priv;
			return GBM_ERR_BMP_PLANES;
		}
		if ( cBitCount != 1 && cBitCount != 4 && cBitCount != 8 && cBitCount != 24 )
		{
			delete bmp_priv;
			return GBM_ERR_BMP_BITCOUNT;
		}
		
		pImage->iWidth   = (int) ulWidth;
		pImage->iHeight   = (int) ulHeight;
		pImage->iBitPerPixel = (int) cBitCount;
		
		bmp_priv->windows       = TRUE;
		bmp_priv->cbFix         = cbFix;
		bmp_priv->ulCompression = ulCompression;
		bmp_priv->cclrUsed      = cclrUsed;
	}
	else
	{
		delete bmp_priv;
		return GBM_ERR_BMP_CBFIX;
	}
	
	*HeaderInfo = bmp_priv;
	pImage->dwFlags = FORMAT_BGR | IMAGE_BMP ;

	return GBM_ERR_OK;
}

DLL_FUNC int ReadBMPData(FILE_HANDLE file, ImageStruct* pImage, void* HeaderInfo, void* pData)
{
	if (!file || !pImage || !HeaderInfo || !pData)
	{
		return GBM_ERR_BAD_ARG;
	}
	
	BMP_PRIV *bmp_priv = (BMP_PRIV *)HeaderInfo ;
	
	byte* data = (byte*)pData;

	int cLinesWorth = ((pImage->iBitPerPixel * pImage->iWidth + 31) >> 5) << 2;
	
	if ( bmp_priv->windows )
	{
		file_seek(file, (long)bmp_priv->offBits, SEEK_SET);
		
		switch ( (int) bmp_priv->ulCompression )
		{
		case BCA_UNCOMP:
			file_read(file, data, pImage->iHeight * cLinesWorth);
			break;
		case BCA_RLE8:
			{
				AHEAD *ahead;
				int x = 0, y = 0;
				BOOLEAN eof8 = FALSE;
				
				if ( (ahead = create_ahead(file)) == NULL )
				{
					delete bmp_priv;
					return GBM_ERR_MEM;
				}
				
				while ( !eof8 )
				{
					byte c = (byte) read_ahead(ahead);
					byte d = (byte) read_ahead(ahead);
					
					if ( c )
					{
						memset(data, d, c);
						x += c;
						data += c;
					}
					else
						switch ( d )
					{
					case MSWCC_EOL:
					{
						int to_eol = cLinesWorth - x;
						
						memset(data, 0, (size_t) to_eol);
						data += to_eol;
						x = 0;
						if ( ++y == pImage->iHeight )
							eof8 = TRUE;
					}
					break;
					case MSWCC_EOB:
						if ( y < pImage->iHeight )
						{
							int to_eol = cLinesWorth - x;
							
							memset(data, 0, (size_t) to_eol);
							x = 0; y++;
							data += to_eol;
							while ( y < pImage->iHeight )
							{
								memset(data, 0, (size_t) cLinesWorth);
								data += cLinesWorth;
								y++;
							}
						}
						eof8 = TRUE;
						break;
					case MSWCC_DELTA:
						{
							byte dx = (byte) read_ahead(ahead);
							byte dy = (byte) read_ahead(ahead);
							int fill = dx + dy * cLinesWorth;
							memset(data, 0, (size_t) fill);
							data += fill;
							x += dx; y += dy;
							if ( y == pImage->iHeight )
								eof8 = TRUE;
						}
						break;
					default:
						{
							int n = (int) d;
							
							while ( n-- > 0 )
								*data++ = (byte) read_ahead(ahead);
							x += d;
							if ( d & 1 )
								read_ahead(ahead); /* Align */
						}
						break;
					}
				}
				
				destroy_ahead(ahead);
			}
			break;
		case BCA_RLE4:
			{
				AHEAD *ahead;
				int x = 0, y = 0;
				BOOLEAN eof4 = FALSE;
				int inx = 0;
				
				if ( (ahead = create_ahead(file)) == NULL )
				{
					delete bmp_priv;
					return GBM_ERR_MEM;
				}
				
				memset(data, 0, (size_t) (pImage->iHeight * cLinesWorth));
				
				while ( !eof4 )
				{
					byte c = (byte) read_ahead(ahead);
					byte d = (byte) read_ahead(ahead);
					
					if ( c )
					{
						byte h, l;
						int i;
						if ( x & 1 )
						{ h = (byte) (d >> 4); l = (byte) (d << 4); }
						else
						{ h = (byte) (d&0xf0); l = (byte) (d&0x0f); }
						for ( i = 0; i < (int) c; i++, x++ )
						{
							if ( x & 1U )
								data[inx++] |= l;
							else
								data[inx]   |= h;
						}
					}
					else
						switch ( d )
					{
					case MSWCC_EOL:
						x = 0;
						if ( ++y == pImage->iHeight )
						eof4 = TRUE;
						inx = cLinesWorth * y;
						break;
					case MSWCC_EOB:
						eof4 = TRUE;
						break;
					case MSWCC_DELTA:
					{
						byte dx = (byte) read_ahead(ahead);
						byte dy = (byte) read_ahead(ahead);
				
						x += dx; y += dy;
						inx = y * cLinesWorth + (x/2);
				
						if ( y == pImage->iHeight )
							eof4 = TRUE;
					}
					break;
					default:
					{
						int i, nr = 0;
				
						if ( x & 1 )
						{
							for ( i = 0; i+2 <= (int) d; i += 2 )
							{
								byte b = (byte) read_ahead(ahead);
								data[inx++] |= (b >> 4);
								data[inx  ] |= (b << 4);
								nr++;
							}
							if ( i < (int) d )
							{
								data[inx++] |= ((byte) read_ahead(ahead) >> 4);
								nr++;
							}
						}
						else
						{
							for ( i = 0; i+2 <= (int) d; i += 2 )
							{
								data[inx++] = (byte) read_ahead(ahead);
								nr++;
							}
							if ( i < (int) d )
							{
								data[inx] = (byte) read_ahead(ahead);
								nr++;
							}
						}
						x += d;
						
						if ( nr & 1 )
							read_ahead(ahead); /* Align input stream to next word */
					}
					break;
				}
			}
				
			destroy_ahead(ahead);
		}
		break;
		default:
			{
				delete bmp_priv;
				return GBM_ERR_BMP_COMP;
			}
		}
	}
	else
	{
		file_seek(file, (long) bmp_priv->offBits, SEEK_SET);
		file_read(file, data, cLinesWorth * pImage->iHeight);
	}

	return GBM_ERR_OK;
}

DLL_FUNC int ReadBMPPalette(FILE_HANDLE file, ImageStruct* pImage, void* HeaderInfo, PALETTE* pPalette)
{
	if (!file || !pImage || !HeaderInfo || !pPalette)
	{
		return GBM_ERR_BAD_ARG;
	}
	
	BMP_PRIV *bmp_priv = (BMP_PRIV *) HeaderInfo;
		
	if (pImage->iBitPerPixel != 24 )
	{
		int i;
		byte b[4];
			
		if ( bmp_priv->windows )
		{
			file_seek(file, (long) (bmp_priv->base + 14L + bmp_priv->cbFix), SEEK_SET);
			pPalette->colorcount = (1 << pImage->iBitPerPixel);
			for ( i = 0; i < pPalette->colorcount; i++ )
				/* Used to use bmp_priv->cclrUsed, but bitmaps have been found which have
				their used colours not at the start of the palette. */
			{
				file_read(file, b, 4);
				pPalette->colors[i].r = b[0];
				pPalette->colors[i].g = b[1];
				pPalette->colors[i].b = b[2];
			}
		}
		else
		{
			file_seek(file, (long) (bmp_priv->base + 26L), SEEK_SET);
			pPalette->colorcount = (1 << pImage->iBitPerPixel);
			for ( i = 0; i < pPalette->colorcount; i++ )
			{
				file_read(file, b, 3);
				pPalette->colors[i].r = b[0];
				pPalette->colors[i].g = b[1];
				pPalette->colors[i].b = b[2];
			}
		}
	}

	return GBM_ERR_OK;
}

DLL_FUNC void DeleteBMPHeaderInfo(void* HeaderInfo)
{
	delete (BMP_PRIV*)HeaderInfo;
}