#include "gif.h"
#include "gbio.h"
#include "translate.h"

//********************************************************

static int step_ilace(int y, int h, int *pass)
{
	switch ( *pass )
	{
	case 0: y += 8; break;
	case 1: y += 8; break;
	case 2: y += 4; break;
	case 3: y += 2; break;
	}

	if ( y < h ) return y;
	if ( *pass == 0 ) 
	{
		*pass = 1; y = 4; if ( y < h ) return y; 
	}
	if ( *pass == 1 ) 
	{
		*pass = 2; y = 2; if ( y < h ) return y; 
	}
	if ( *pass == 2 ) 
	{
		*pass = 3; y = 1; 
	}

	return y;
}

#define low_byte(w) ((byte)  (          (w)&0x00ffU)    )
#define high_byte(w)  ((byte) (((unsigned)(w)&0xff00U)>>8))
#define make_word(a,b)  (((word)a) + (((word)b) << 8))

struct GIF_PRIV
{
	BOOLEAN ilace;
	int bpp;
	byte pal[0x100*3];
};

typedef unsigned int cword;

struct READ_CONTEXT
{
	FILE_HANDLE file;   /* File descriptor to read */
	int inx, size;      /* Index and size in bits */
	byte buf[255+3];    /* Buffer holding bits */
	int code_size;      /* Number of bits to return at once */
	cword read_mask;    /* 2^code_size-1 */
} ;

static cword read_code(READ_CONTEXT *c)
{
	dword raw_code; int byte_inx;
	
	while ( c->inx + c->code_size > c->size )
		/*...snot enough bits in buffer\44\ refill it:16:*/
		/* Not very efficient, but infrequently called */
		
	{
		int bytes_to_lose = ((unsigned) c->inx >> 3);
		byte bytes;
		
		/* Note biggest code size is 12 bits */
		/* And this can at worst span 3 bytes */
		memcpy(c->buf, c->buf + bytes_to_lose, 3);
		(c->inx) &= 7;
		(c->size) -= (bytes_to_lose << 3);
		if ( file_read(c->file, &bytes, 1) != 1 )
			return 0xffff;
		if ( file_read(c->file, c->buf + ((unsigned) c->size >> 3), bytes) != bytes )
			return 0xffff;
		(c->size) += (bytes << 3);
	}
	
	byte_inx = ((unsigned) c->inx >> 3);
	raw_code = c->buf[byte_inx] + ((c->buf[byte_inx + 1]) << 8);
	if ( c->code_size > 8 )
		raw_code += ((c->buf[byte_inx + 2]) << 16);
	raw_code >>= ((c->inx) & 7U);
	(c->inx) += (byte) (c->code_size);
	
	return (cword) raw_code & c->read_mask;
}

struct OUTPUT_CONTEXT
{
	int x, y, w, h, bpp, pass;
	BOOLEAN ilace;
	int stride;
	byte *data, *data_this_line;
};

static void output(byte value, OUTPUT_CONTEXT *o)
{
	if ( o->y >= o->h )
		return;
	
	switch ( o->bpp )
	{
	case 1:
		if ( (o->x) & 7U )
			o->data_this_line[(unsigned)(o->x) >> 3] |= (value << (7 - (((unsigned)o->x) & 7U)));
		else
			o->data_this_line[(unsigned)(o->x) >> 3] = (value << 7);
		break;
	case 4:
		if ( (o->x) & 1U )
			o->data_this_line[(unsigned)(o->x) >> 1] |= value;
		else
			o->data_this_line[(unsigned)(o->x) >> 1] = (value << 4);
		break;
	case 8:
		o->data_this_line[o->x] = value;
		break;
	}
	
	if ( ++(o->x) < o->w )
		return;
	
	o->x = 0;
	if ( o->ilace )
	{
		o->y = step_ilace(o->y, o->h, &(o->pass));
		o->data_this_line = o->data + (o->h - 1 - o->y) * o->stride;
	}
	else
	{
		(o->y)++;
		o->data_this_line -= (o->stride);
	}
}

//********************************************************


DLL_FUNC int ReadGIFHeader(FILE_HANDLE file, ImageStruct* pImage, void** HeaderInfo)
{
	if (!file || !pImage || ! HeaderInfo)
	{
		return GBM_ERR_BAD_ARG;
	}
	
	GIF_PRIV *gif_priv = new GIF_PRIV;
	if (!gif_priv)
		return GBM_ERR_MEM;
	*HeaderInfo = NULL;

	byte signiture[6], scn_desc[7], image_desc[10];
	int img = -1, img_want = pImage->Options;//img_want - number of image to load
	int bits_gct;
	
	/* Read and validate signiture block */
	
	if ( file_read(file, signiture, 6) != 6 )
	{
		delete gif_priv;
		return GBM_ERR_READ;
	}
	if ( memcmp(signiture, "GIF87a", 6) &&
		memcmp(signiture, "GIF89a", 6) )
	{
		delete gif_priv;
		return GBM_ERR_BAD_MAGIC;
	}
	
	/* Read screen descriptor */
	
	if ( file_read(file, scn_desc, 7) != 7 )
	{
		delete gif_priv;
		return GBM_ERR_READ;
	}
	gif_priv->bpp = bits_gct = (scn_desc[4] & 7) + 1;
	
	if ( scn_desc[4] & 0x80 )
		/* Global colour table follows screen descriptor */
	{
		if ( file_read(file, gif_priv->pal, 3 << bits_gct) != (3 << bits_gct) )
		{
			delete gif_priv;
			return GBM_ERR_READ;
		}
	}
	else
		/* Blank out palette, but make entry 1 white */
	{
		memset(gif_priv->pal, 0, 3 << bits_gct);
		gif_priv->pal[3] =
			gif_priv->pal[4] =
			gif_priv->pal[5] = 0xff;
	}
	
	/* Expected image descriptors / extension blocks / terminator */
	
	while ( img < img_want )
	{
		if ( file_read(file, image_desc, 1) != 1 )
		{
			delete gif_priv;
			return GBM_ERR_READ;
		}
		switch ( image_desc[0] )
		{
		case 0x2c:
			if ( file_read(file, image_desc + 1, 9) != 9 )
			{
				delete gif_priv;
				return GBM_ERR_READ;
			}
			pImage->iWidth = make_word(image_desc[5], image_desc[6]);
			pImage->iHeight = make_word(image_desc[7], image_desc[8]);
			gif_priv->ilace = ( (image_desc[9] & 0x40) != 0 );
			
			if ( image_desc[9] & 0x80 )
				/* Local colour table follows */
			{
				gif_priv->bpp = (image_desc[9] & 7) + 1;
				if ( file_read(file, gif_priv->pal, 3 << gif_priv->bpp) != (3 << gif_priv->bpp) )
				{
					delete gif_priv;
					return GBM_ERR_READ;
				}
			}
			
			if ( ++img != img_want )
				/* Skip the image data */
			{
				byte code_size, block_size;
				
				if ( file_read(file, &code_size, 1) != 1 )
				{
					delete gif_priv;
					return GBM_ERR_READ;
				}
				do
				{
					if ( file_read(file, &block_size, 1) != 1 )
						return GBM_ERR_READ;
					file_seek(file, block_size, SEEK_CUR);
				}
				while ( block_size );
			}
			
			break;
			
		case 0x21:
			{
				byte func_code, byte_count;
				
				if ( file_read(file, &func_code, 1) != 1 )
				{
					delete gif_priv;
					return GBM_ERR_READ;
				}
				do
				{
					if ( file_read(file, &byte_count, 1) != 1 )
					{
						delete gif_priv;
						return GBM_ERR_READ;
					}
					file_seek(file, byte_count, SEEK_CUR);
				}
				while ( byte_count );
			}
			break;
			
		case 0x3b:
			{
				delete gif_priv;
				return GBM_ERR_GIF_TERM;
			}
		default:
			{
				delete gif_priv;
				return GBM_ERR_GIF_HEADER;
			}
		}
	}
	
	switch ( gif_priv->bpp )
	{
	case 1:   pImage->iBitPerPixel = 1;   break;
	case 2:
	case 3:
	case 4:   pImage->iBitPerPixel = 4;   break;
	case 5:
	case 6:
	case 7:
	case 8:   pImage->iBitPerPixel = 8;   break;
	default:  
		return GBM_ERR_GIF_BPP;
	}

	*HeaderInfo = gif_priv;
	pImage->dwFlags = FORMAT_RGB | IMAGE_GIF;

	return GBM_ERR_OK;
}

DLL_FUNC int ReadGIFData(FILE_HANDLE file, ImageStruct* pImage, void* HeaderInfo, void* pData)
{
	if (!file || !pImage || !HeaderInfo || !pData)
	{
		return GBM_ERR_BAD_ARG;
	}

	GIF_PRIV *gif_priv = (GIF_PRIV *) HeaderInfo;
	
	byte* data = (byte*)pData;

	byte min_code_size;   /* As read from the file */
	int init_code_size;   /* Initial code size */
	cword max_code;     /* 1 << code_size */
	cword clear_code;   /* Code to clear table */
	cword eoi_code;     /* End of information code */
	cword first_free_code;    /* First free code */
	cword free_code;    /* Next available free code slot */
	word bit_mask;      /* Output pixel mask */ 
	int i, out_count = 0;
	cword code, cur_code, old_code, in_code, fin_char;
	cword *prefix, *suffix, *outcode;
	READ_CONTEXT c;
	OUTPUT_CONTEXT o;
	BOOLEAN table_full = FALSE; /* To help implement deferred clear */
	
	if ( (prefix = (cword *) malloc((size_t) (4096 * sizeof(cword)))) == NULL )
	{
		delete gif_priv;
		return GBM_ERR_MEM;
	}
	if ( (suffix = (cword *) malloc((size_t) (4096 * sizeof(cword)))) == NULL )
	{
		free(prefix);
		delete gif_priv;
		return GBM_ERR_MEM;
	}
	if ( (outcode = (cword *) malloc((size_t) (4097 * sizeof(cword)))) == NULL )
	{
		free(suffix);   
		free(prefix);
		delete gif_priv;
		return GBM_ERR_MEM;
	}
	
	if ( file_read(file, &min_code_size, 1) != 1 )
	{ 
		free(outcode);
		free(suffix);
		free(prefix);
		delete gif_priv;
		return GBM_ERR_READ;
	}
	
	if ( min_code_size < 2 || min_code_size > 9 )
	{ 
		free(outcode);
		free(suffix);
		free(prefix);
		delete gif_priv;
		return GBM_ERR_GIF_CODE_SIZE;
	}
	
	/* Initial read context */
	
	c.inx            = 0;
	c.size           = 0;
	c.file             = file;
	c.code_size      = min_code_size + 1;
	c.read_mask      = (cword) (( 1 << c.code_size ) - 1);
	
	/* Initialise pixel-output context */
	
	o.x              = 0;
	o.y              = 0;
	o.pass           = 0;
	o.w              = pImage->iWidth;
	o.h              = pImage->iHeight;
	o.bpp            = pImage->iBitPerPixel;
	o.ilace          = gif_priv->ilace;
	o.stride         = ( (pImage->iWidth * pImage->iBitPerPixel + 31) / 32 ) * 4;
	o.data           = data;
	o.data_this_line = data + (pImage->iHeight - 1) * o.stride;
	
	bit_mask = (word) ((1 << gif_priv->bpp) - 1);
	
	/* 2^min_code size accounts for all colours in file */
	
	clear_code = (cword) ( 1 << min_code_size );
	eoi_code = (cword) (clear_code + 1);
	free_code = first_free_code = (cword) (clear_code + 2);
	
	/* 2^(min_code_size+1) includes clear and eoi code and space too */
	
	init_code_size = c.code_size;
	max_code = (cword) ( 1 << c.code_size );
	
	while ( o.y < o.h && (code = read_code(&c)) != eoi_code && code != 0xffff )
	{
		if ( code == clear_code )
		{
			c.code_size = init_code_size;
			max_code = (cword) ( 1 << c.code_size );
			c.read_mask = (cword) (max_code - 1);
			free_code = first_free_code;
			cur_code = old_code = code = read_code(&c);
			if ( code == 0xffff )
				break;
			fin_char = (cur_code & bit_mask);
			output((byte) fin_char, &o);
			table_full = FALSE;
		}
		else
		{
			cur_code = in_code = code;
			if ( cur_code >= free_code )
			{
				cur_code = old_code;
				outcode[out_count++] = fin_char;
			}
			while ( cur_code > bit_mask )
			{
				if ( out_count > 4096 )
				{
					free(outcode);
					free(suffix);
					free(prefix);
					delete gif_priv;
					return GBM_ERR_GIF_CORRUPT;
				}
				outcode[out_count++] = suffix[cur_code];
				cur_code = prefix[cur_code];
			}
			fin_char = (cur_code & bit_mask);
			outcode[out_count++] = fin_char;
			for ( i = out_count - 1; i >= 0; i-- )
				output((byte) outcode[i], &o);
			out_count = 0;
			
			/* Update dictionary */
			
			if ( !table_full )
			{
				prefix[free_code] = old_code;
				suffix[free_code] = fin_char;
				
				/* Advance to next free slot */
				
				if ( ++free_code >= max_code )
				{
					if ( c.code_size < 12 )
					{
						c.code_size++;
						max_code <<= 1;
						c.read_mask = (cword) (( 1 << c.code_size ) - 1);
					}
					else
						table_full = TRUE;
				}
			}
			old_code = in_code;
		}
	}
	
	free(outcode);
	free(suffix);
	free(prefix);
	delete gif_priv;
	
	if ( o.y < o.h && code == 0xffff )
		/* If ran out of data and hadn't got to the end */
		return GBM_ERR_READ;
	
	return GBM_ERR_OK;
}

DLL_FUNC int ReadGIFPalette(FILE_HANDLE file, ImageStruct* pImage, void* HeaderInfo, PALETTE* pPalette)
{
	if (!file || !pImage || !HeaderInfo || !pPalette)
	{
		return GBM_ERR_BAD_ARG;
	}

	GIF_PRIV *gif_priv = (GIF_PRIV *) HeaderInfo;
	if (!gif_priv)
	{
		return GBM_ERR_BAD_ARG;
	}
	byte *pal = gif_priv->pal;
	int i;
	
	memset(pPalette->colors, 0x80, sizeof(pPalette->colors));
	
	pPalette->colorcount = (1 << gif_priv->bpp);
	for ( i = 0; i < (1 << gif_priv->bpp); i++ )
	{
		pPalette->colors[i].r = *pal++;
		pPalette->colors[i].g = *pal++;
		pPalette->colors[i].b = *pal++;
	}
	
	return GBM_ERR_OK;
}

HBITMAP _LoadGIFFile(FILE_HANDLE file, ImageStruct* pImage)
{
	void* HeaderInfo;
	HBITMAP Bitmap;

	if (ReadGIFHeader(file,pImage,&HeaderInfo) != GBM_ERR_OK)
	{
		file_close(file);
		return NULL;
	}

	if (pImage->iBitPerPixel > 8)
	{
		file_close(file);
		return NULL;
	}

	PALETTE pal;
	if (ReadGIFPalette(file,pImage,HeaderInfo,&pal) != GBM_ERR_OK)
	{
		file_close(file);
		return NULL;
	}

	unsigned char* pBits;
	int ColorCount = 1 << pImage->iBitPerPixel;
	BITMAPINFO* pbmInfo = (BITMAPINFO*)malloc(sizeof(BITMAPINFO) + (ColorCount - 1) * sizeof(RGBQUAD));
	if (!pbmInfo)
	{
		file_close(file);
		return NULL;
	}
	BITMAPINFOHEADER* pBitmapHeader = &pbmInfo->bmiHeader;
	pBitmapHeader->biBitCount = pImage->iBitPerPixel;
	pBitmapHeader->biClrImportant = 0;
	pBitmapHeader->biClrUsed = 0;
	pBitmapHeader->biCompression = BI_RGB;
	pBitmapHeader->biHeight = (int)pImage->iHeight;
	pBitmapHeader->biPlanes = 1;
	pBitmapHeader->biSize = sizeof(BITMAPINFOHEADER);
	pBitmapHeader->biSizeImage = 0;
	pBitmapHeader->biWidth = pImage->iWidth;
	pBitmapHeader->biXPelsPerMeter = 0;
	pBitmapHeader->biYPelsPerMeter = 0;
	for (int i = 0; i < ColorCount; i++)
	{
		pbmInfo->bmiColors[i].rgbBlue = pal.colors[i].b;
		pbmInfo->bmiColors[i].rgbGreen = pal.colors[i].g;
		pbmInfo->bmiColors[i].rgbRed = pal.colors[i].r;
		pbmInfo->bmiColors[i].rgbReserved = 0;
	}
	HDC dc = GetDC(0);
	Bitmap = CreateDIBSection(dc, pbmInfo, DIB_RGB_COLORS, (void**)&pBits, NULL, 0);
	if (!Bitmap)
	{
		free(pbmInfo);
		file_close(file);
		return NULL;
	}
	if (ReadGIFData(file, pImage, HeaderInfo, pBits) != GBM_ERR_OK)
	{
		file_close(file);
		DeleteObject(Bitmap);
		return NULL;
	}
	free(pbmInfo);

	return Bitmap;
}

DLL_FUNC HBITMAP LoadGIFFileA(char* filename,ImageStruct* pImage)
{
	FILE_HANDLE file = file_openA(filename);
	if (file == (FILE_HANDLE)-1)
		return NULL;
	
	return _LoadGIFFile(file, pImage);
}

DLL_FUNC HBITMAP LoadGIFFileW(wchar_t* filename,ImageStruct* pImage)
{
	FILE_HANDLE file = file_openW(filename);
	if (file == (FILE_HANDLE)-1)
		return NULL;

	return _LoadGIFFile(file, pImage);
}

DLL_FUNC void DeleteGIFHeaderInfo(void* HeaderInfo)
{
	delete (GIF_PRIV*)HeaderInfo;
}
