#include "png.h"
#include "png\pnglib.h"
#include "gbio.h"
#include "translate.h"
//******************************************************************

#ifndef png_jmpbuf
#  define png_jmpbuf(png_ptr) ((png_ptr)->jmpbuf)
#endif

#define PNG_BYTES_TO_CHECK 4
int check_if_png(FILE_HANDLE file)
{
	unsigned char buf[PNG_BYTES_TO_CHECK];
	
	/* Read in some of the signature bytes */
	if (file_read(file, buf, PNG_BYTES_TO_CHECK) != PNG_BYTES_TO_CHECK)
		return 0;
	
		/* Compare the first PNG_BYTES_TO_CHECK bytes of the signature.
	Return nonzero (true) if they match */
	
	file_seek(file, 0, SEEK_SET);

	return(!png_sig_cmp(buf, (png_size_t)0, PNG_BYTES_TO_CHECK));
}

void mypngread(png_structp png_ptr, unsigned char* pbuf, unsigned int size)
{
	if (file_read((FILE_HANDLE)png_ptr->io_ptr,pbuf, size) != size)
	{
		png_error(png_ptr, "read error");
	}
}

struct PNG_PRIV 
{
	png_structp png_ptr;
	png_infop info_ptr;
};
//******************************************************************

DLL_FUNC int ReadPNGHeader(FILE_HANDLE file, ImageStruct* pImage, void** HeaderInfo)
{
	unsigned int sig_read = 0;
	
	if (!file || !pImage || !HeaderInfo)
	{
		return GBM_ERR_BAD_ARG;
	}
	
	if (!check_if_png(file))
	{
		return GBM_ERR_BAD_MAGIC;
	}

	PNG_PRIV* priv = new PNG_PRIV;
	if (!priv)
	{
		return GBM_ERR_MEM;
	}

	*HeaderInfo = NULL;
	priv->png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	if (priv->png_ptr == NULL)
	{
		return GBM_ERR_PNG_INIT;
	}

	priv->info_ptr = png_create_info_struct(priv->png_ptr);
	if (priv->info_ptr == NULL)
	{
		png_destroy_read_struct(&priv->png_ptr, (png_infopp)NULL, (png_infopp)NULL);
		return GBM_ERR_MEM;
	}

	if (setjmp(png_jmpbuf(priv->png_ptr)))
	{
		/* Free all of the memory associated with the png_ptr and info_ptr */
		png_destroy_read_struct(&priv->png_ptr, &priv->info_ptr, (png_infopp)NULL);
		/* If we get here, we had a problem reading the file */
		return GBM_ERR_PNG_GENERAL_ERROR;
	}

	png_set_read_fn(priv->png_ptr, (void *)file, mypngread);
	png_read_info(priv->png_ptr, priv->info_ptr);

	pImage->iWidth = priv->info_ptr->width;
	pImage->iHeight = priv->info_ptr->height;
	
	int colortype = priv->info_ptr->color_type;
	int channels ;
	if (colortype == PNG_COLOR_TYPE_PALETTE)
		channels = 1;
	else
		if (colortype & PNG_COLOR_MASK_COLOR)
			channels = 3;
		else
			channels = 1;
	if (colortype & PNG_COLOR_MASK_ALPHA)
		channels++;

	int bit_depth = priv->info_ptr->bit_depth;
	if (bit_depth == 16)
		bit_depth = 8;
	pImage->iBitPerPixel = (channels * bit_depth);
	if (pImage->iBitPerPixel <= 24)
		pImage->dwFlags = FORMAT_RGB | IMAGE_PNG | OPTIONS_FLIP;
	if (pImage->iBitPerPixel == 32)
	{
		pImage->dwFlags = FORMAT_RGBA | IMAGE_PNG | OPTIONS_FLIP;
	}
	*HeaderInfo = priv;

	return GBM_ERR_OK;
}

DLL_FUNC int ReadPNGData(FILE_HANDLE file, ImageStruct* pImage, void* HeaderInfo, void* pData)
{
	png_uint_32 i,image_height;
	int pass, j;
	
	if (!file || !pImage || !HeaderInfo || !pData)
	{
		return GBM_ERR_BAD_ARG;
	}

	PNG_PRIV* priv = (PNG_PRIV*)HeaderInfo;

	if (setjmp(png_jmpbuf(priv->png_ptr)))
	{
		/* Free all of the memory associated with the png_ptr and info_ptr */
		png_destroy_read_struct(&priv->png_ptr, &priv->info_ptr, (png_infopp)NULL);
		/* If we get here, we had a problem reading the file */
		return GBM_ERR_PNG_GENERAL_ERROR;
	}

	png_set_strip_16(priv->png_ptr);

	png_debug(1, "in png_read_image\n");
	/* save jump buffer and error functions */
	
#ifdef PNG_READ_INTERLACING_SUPPORTED
	pass = png_set_interlace_handling(priv->png_ptr);
#else
	if (priv->png_ptr->interlaced)
		png_error(priv->png_ptr,
        "Cannot read interlaced image -- interlace handler disabled.");
	pass = 1;
#endif
	
	image_height=priv->png_ptr->height;
	priv->png_ptr->num_rows = image_height; /* Make sure this is set correctly */
	
	unsigned char* data = (unsigned char*)pData;
	int stride = ((pImage->iWidth * pImage->iBitPerPixel +31) >> 5) <<2;

	for (j = 0; j < pass; j++)
	{
		for (i = 0; i < image_height; i++)
		{
			png_read_row(priv->png_ptr, data, NULL);
			data += stride;
		}
	}

	png_read_end(priv->png_ptr, priv->info_ptr);
	
	png_destroy_read_struct(&priv->png_ptr, &priv->info_ptr, (png_infopp)NULL);

	delete priv;
	
	return GBM_ERR_OK;
}

DLL_FUNC int ReadPNGPalette(FILE_HANDLE file, ImageStruct* pImage, void* HeaderInfo, PALETTE* pPalette)
{
	if (!file || !pImage || !HeaderInfo || !pPalette)
	{
		return GBM_ERR_BAD_ARG;
	}

	PNG_PRIV* priv = (PNG_PRIV*)HeaderInfo;
	if (priv->info_ptr->color_type == PNG_COLOR_TYPE_PALETTE)
	{
		pPalette->colorcount = priv->info_ptr->num_palette;
		for (int i = 0; i < priv->info_ptr->num_palette; i++)
		{
			pPalette->colors[i].b = priv->info_ptr->palette[i].blue;
			pPalette->colors[i].g = priv->info_ptr->palette[i].green;
			pPalette->colors[i].r = priv->info_ptr->palette[i].red;
		}
	}
	else
	{
		pPalette->colorcount = (1 << pImage->iBitPerPixel);
		int step = 255 / (pPalette->colorcount - 1);
		for(int i = 0; i < 256; i += step)
		{
			pPalette->colors[i].b = pPalette->colors[i].g = pPalette->colors[i].r = i;
		}
	}

	return GBM_ERR_OK;
}

HBITMAP _LoadPNGFile(FILE_HANDLE file, ImageStruct* pImage)
{
	void* HeaderInfo;
	HBITMAP Bitmap;

	if (ReadPNGHeader(file,pImage,&HeaderInfo) != GBM_ERR_OK)
	{
		file_close(file);
		return NULL;
	}

	if (pImage->iBitPerPixel > 8)
	{
		BITMAPINFO bmInfo;
		unsigned char* pBits;
		BITMAPINFOHEADER* pBitmapHeader = &bmInfo.bmiHeader;
		pBitmapHeader->biBitCount = pImage->iBitPerPixel;
		pBitmapHeader->biClrImportant = 0;
		pBitmapHeader->biClrUsed = 0;
		pBitmapHeader->biCompression = BI_RGB;
		pBitmapHeader->biHeight = -(int)pImage->iHeight;
		pBitmapHeader->biPlanes = 1;
		pBitmapHeader->biSize = sizeof(BITMAPINFOHEADER);
		pBitmapHeader->biSizeImage = 0;
		pBitmapHeader->biWidth = pImage->iWidth;
		pBitmapHeader->biXPelsPerMeter = 0;
		pBitmapHeader->biYPelsPerMeter = 0;
		HDC dc = GetDC(0);
		Bitmap=CreateDIBSection(dc,&bmInfo,DIB_RGB_COLORS,(void**)&pBits,NULL,0);
		if (!Bitmap)
		{
			file_close(file);
			return NULL;
		}
		if (ReadPNGData(file, pImage, HeaderInfo,pBits) != GBM_ERR_OK)
		{
			file_close(file);
			DeleteObject(Bitmap);
			return NULL;
		}

		if (pImage->iBitPerPixel == 24)
		{
			RGBtoBGR(pBits,pImage);
		}

		if (pImage->iBitPerPixel == 32)
		{
			RGBAtoBGRA(pBits,pImage);
		}
	}
	else
	{
		PALETTE pal;
		if (ReadPNGPalette(file,pImage,HeaderInfo,&pal) != GBM_ERR_OK)
		{
			file_close(file);
			return NULL;
		}

		unsigned char* pBits;
		int ColorCount = 1 << pImage->iBitPerPixel;
		BITMAPINFO* pbmInfo = (BITMAPINFO*)malloc(sizeof(BITMAPINFO) + (ColorCount - 1) * sizeof(RGBQUAD));
		if (!pbmInfo)
		{
			file_close(file);
			return NULL;
		}
		BITMAPINFOHEADER* pBitmapHeader = &pbmInfo->bmiHeader;
		pBitmapHeader->biBitCount = pImage->iBitPerPixel;
		pBitmapHeader->biClrImportant = 0;
		pBitmapHeader->biClrUsed = 0;
		pBitmapHeader->biCompression = BI_RGB;
		pBitmapHeader->biHeight = -(int)pImage->iHeight;
		pBitmapHeader->biPlanes = 1;
		pBitmapHeader->biSize = sizeof(BITMAPINFOHEADER);
		pBitmapHeader->biSizeImage = 0;
		pBitmapHeader->biWidth = pImage->iWidth;
		pBitmapHeader->biXPelsPerMeter = 0;
		pBitmapHeader->biYPelsPerMeter = 0;
		for (int i = 0; i < ColorCount; i++)
		{
			pbmInfo->bmiColors[i].rgbBlue = pal.colors[i].b;
			pbmInfo->bmiColors[i].rgbGreen = pal.colors[i].g;
			pbmInfo->bmiColors[i].rgbRed = pal.colors[i].r;
			pbmInfo->bmiColors[i].rgbReserved = 0;
		}
		HDC dc = GetDC(0);
		Bitmap = CreateDIBSection(dc, pbmInfo, DIB_RGB_COLORS, (void**)&pBits, NULL, 0);
		if (!Bitmap)
		{
			free(pbmInfo);
			file_close(file);
			return NULL;
		}
		if (ReadPNGData(file, pImage, HeaderInfo, pBits) != GBM_ERR_OK)
		{
			file_close(file);
			DeleteObject(Bitmap);
			return NULL;
		}
		free(pbmInfo);

	}

	return Bitmap;
}

DLL_FUNC HBITMAP LoadPNGFileA(char* filename,ImageStruct* pImage)
{
	FILE_HANDLE file = file_openA(filename);
	if (file == (FILE_HANDLE)-1)
		return NULL;
	
	return _LoadPNGFile(file, pImage);
}

DLL_FUNC HBITMAP LoadPNGFileW(wchar_t* filename,ImageStruct* pImage)
{
	FILE_HANDLE file = file_openW(filename);
	if (file == (FILE_HANDLE)-1)
		return NULL;

	return _LoadPNGFile(file, pImage);
}

DLL_FUNC void DeletePNGHeaderInfo(void* HeaderInfo)
{
	delete (PNG_PRIV*)HeaderInfo;
}