#include "translate.h"
#include <MEMORY.H>
#include <XUTILITY>

int PaletteToRGB(byte* source, byte* dest, PALETTE* pPalette, ImageStruct* pImage)
{
	int lineshift = (-3*pImage->iWidth) & 0x00000003;
	int lsh2 = (-pImage->iWidth * pImage->iBitPerPixel >> 3) & 0x3;
	switch(pImage->iBitPerPixel) 
	{
	case 1:
		{
			byte register data = 0;
			int register offset;
			int c;
			for (int i = 0; i < pImage->iHeight; i++)
			{
				offset = 0;
				for (int j = 0; j < pImage->iWidth; j++)
				{
					if (!offset)
					{
						data = *source;
						source += 1;
						offset = 8;
					}
					
					offset--;
					if (data & (1 << offset) )
					{
						c =  1;
					}
					else
					{
						c = 0;
					}
					memcpy(dest,&pPalette->colors[c],3);
					dest += 3;
				}
				dest += lineshift;
				source += lsh2;
			}
		}
		break;
	case 2:
		{
			byte register data = 0;
			int register offset;
			int c;
			for (int i = 0; i < pImage->iHeight; i++)
			{
				offset = 0;
				for (int j = 0; j < pImage->iWidth; j++)
				{
					if (!offset)
					{
						data = *source;
						source ++;
						offset = 4;
					}
					
					offset--;
					c = (data & (3 << (offset << 1)) ) >> (offset << 1);
					
					memcpy(dest,&pPalette->colors[c],3);
					dest += 3;
				}
				dest += lineshift;
				source += lsh2;
			}
		}
		break;
	case 4:
		{
			byte register data = 0;
			int register offset;
			int c;
			for (int i = 0; i < pImage->iHeight; i++)
			{
				offset = 0;
				for (int j = 0; j < pImage->iWidth; j++)
				{
					if (!offset)
					{
						data = *source;
						source ++;
						offset = 2;
					}
					
					offset--;
					c = (data & (15 << (offset << 2)) ) >> (offset << 2);
					
					memcpy(dest,&pPalette->colors[c],3);
					dest += 3;
				}
				dest += lineshift;
				source += lsh2;
			}
		}
		break;
	case 8:
		{
			byte register data = 0;
			for (int i = 0; i < pImage->iHeight; i++)
			{
				for (int j = 0; j < pImage->iWidth; j++)
				{
					data = *source;
					source ++;
										
					memcpy(dest,&pPalette->colors[data],3);
					dest += 3;
				}
				dest += lineshift;
				source += lsh2;
			}
		}
		break;
	default:
		return GBM_ERR_NOT_SUPP;
	}

	return GBM_ERR_OK;
}

void RGBtoBGR(byte* data, ImageStruct* pImage)
{
	int shift = (-pImage->iWidth*3) & 3;
	for (int i = 0; i < pImage->iHeight; i++)
	{
		for (int j = 0; j < pImage->iWidth; j++)
		{
			register byte t = *data;
			*data = *(data+2);
			*(data+2) = t;
			data+=3;
		}
		data += shift;
	}
}

void RGBAtoBGRA(byte* data, ImageStruct* pImage)
{
	for (int i = 0; i < pImage->iHeight * pImage->iWidth; i++)
	{
		register byte t = *data;
		*data = *(data+2);
		*(data+2) = t;
		data+=4;
	}
}

void InvertPalette(PALETTE* pPalette)
{
	for (int i = 0; i < pPalette->colorcount; i++)
	{
		std::swap(pPalette->colors[i].r,pPalette->colors[i].b);
	}

}

int PaletteToRGBA(byte* source, byte* dest, PALETTE* pPalette, ImageStruct* pImage)
{
	//int lineshift = (-3*pImage->iWidth) & 0x00000003;
	int lsh2 = (-pImage->iWidth * pImage->iBitPerPixel >> 3) & 0x3;
	switch(pImage->iBitPerPixel) 
	{
	case 1:
		{
			byte register data = 0;
			int register offset;
			int c;
			for (int i = 0; i < pImage->iHeight; i++)
			{
				offset = 0;
				for (int j = 0; j < pImage->iWidth; j++)
				{
					if (!offset)
					{
						data = *source;
						source += 1;
						offset = 8;
					}
					
					offset--;
					if (data & (1 << offset) )
					{
						c =  1;
					}
					else
					{
						c = 0;
					}
					memcpy(dest,&pPalette->colors[c],3);
					dest += 3;
					*dest = 0xff;
					dest++;
				}
				source += lsh2;
			}
		}
		break;
	case 2:
		{
			byte register data = 0;
			int register offset;
			int c;
			for (int i = 0; i < pImage->iHeight; i++)
			{
				offset = 0;
				for (int j = 0; j < pImage->iWidth; j++)
				{
					if (!offset)
					{
						data = *source;
						source ++;
						offset = 4;
					}
					
					offset--;
					c = (data & (3 << (offset << 1)) ) >> (offset << 1);
					
					memcpy(dest,&pPalette->colors[c],3);
					dest += 3;
					*dest = 0xff;
					dest++;
				}
				source += lsh2;
			}
		}
		break;
	case 4:
		{
			byte register data = 0;
			int register offset;
			int c;
			for (int i = 0; i < pImage->iHeight; i++)
			{
				offset = 0;
				for (int j = 0; j < pImage->iWidth; j++)
				{
					if (!offset)
					{
						data = *source;
						source ++;
						offset = 2;
					}
					
					offset--;
					c = (data & (15 << (offset << 2)) ) >> (offset << 2);
					
					memcpy(dest,&pPalette->colors[c],3);
					dest += 3;
					*dest = 0xff;
					dest++;
				}
				source += lsh2;
			}
		}
		break;
	case 8:
		{
			byte register data = 0;
			for (int i = 0; i < pImage->iHeight; i++)
			{
				for (int j = 0; j < pImage->iWidth; j++)
				{
					data = *source;
					source ++;
										
					memcpy(dest,&pPalette->colors[data],3);
					dest += 3;
					*dest = 0xff;
					dest++;
				}
				source += lsh2;
			}
		}
		break;
	default:
		return GBM_ERR_NOT_SUPP;
	}

	return GBM_ERR_OK;
}